import json
import yaml
from pathlib import Path
from typing import Union


def load_json(file_path: Path) -> Union[dict, list]:
    with open(file_path, 'r') as f:
        return json.load(f)


def save_json(data: Union[dict, list], target_file: str):
    with open(target_file, 'w') as f:
        json.dump(data, f, indent=2, sort_keys=True)


def load_yaml(file_path: str) -> str:
    with open(file_path, 'r') as f:
        return yaml.safe_load(f)



import os
import argparse
from imageatm.components import Training


def run_training(image_dir: str, job_dir: str, **kwargs):
    trainer = Training(image_dir=image_dir, job_dir=job_dir, **kwargs)
    trainer.run()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-j', '--job-dir', help='Directory with job files.', required=True)
    parser.add_argument('-i', '--image-dir', help='Directory with image files.', required=True)

    args = parser.parse_args()

    run_training(**{**os.environ, **args.__dict__})



import numpy as np
from six import iteritems
from multiprocessing import cpu_count
from typing import Tuple, List
from logging import Logger
from tensorflow.python.client import device_lib
from keras.models import load_model as load_model_keras
from keras.callbacks import Callback
from keras.engine.training import Model
from pathlib import Path


def load_model(model_path: Path) -> Model:
    return load_model_keras(str(model_path))


def use_multiprocessing() -> Tuple[bool, int]:
    if _get_available_gpus():
        # if GPU is available, use all available CPUs for batch preprocessing
        use_multiprocessing = True
        num_workers = cpu_count()
    else:
        # device = 'CPU'
        use_multiprocessing = False
        num_workers = 1
    return use_multiprocessing, num_workers


def _get_available_gpus() -> List[str]:
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']


class LoggingMetrics(Callback):
    """Callback for logging metrics at the end of each epoch.

    Args:
        logger: Root logger.
    """

    def __init__(self, logger: Logger) -> None:
        Callback.__init__(self)
        self.logger = logger
        self.format_epoch = 'Epoch: {} - {}'
        self.format_keyvalue = '{}: {:0.4f}'
        self.format_separator = ' - '

    def on_epoch_end(self, epoch: int, logs: dict = {}):
        values = self.format_separator.join(
            self.format_keyvalue.format(k, v) for k, v in iteritems(logs)
        )
        msg = self.format_epoch.format(epoch + 1, values)
        self.logger.debug(msg)


class LoggingModels(Callback):
    def __init__(
        self,
        filepath: Path,
        logger: Logger,
        monitor: str = 'val_loss',
        verbose: int = 0,
        save_best_only: bool = False,
        save_weights_only: bool = False,
        mode: str = 'auto',
        period: int = 1,
    ) -> None:
        self.monitor = monitor
        self.verbose = verbose
        self.filepath = filepath
        self.save_best_only = save_best_only
        self.save_weights_only = save_weights_only
        self.period = period
        self.epochs_since_last_save = 0
        self.logger = logger

        if mode not in ['auto', 'min', 'max']:
            self.logger.warning(
                'ModelCheckpoint mode {} is unknown, fallback to auto mode.'.format(mode)
            )
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
            self.best = np.Inf
        elif mode == 'max':
            self.monitor_op = np.greater
            self.best = -np.Inf
        else:
            if 'acc' in self.monitor or self.monitor.startswith('fmeasure'):
                self.monitor_op = np.greater
                self.best = -np.Inf
            else:
                self.monitor_op = np.less
                self.best = np.Inf

    def on_epoch_end(self, epoch: int, logs: dict = None):
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = Path(str(self.filepath).format(epoch=epoch + 1, **logs))
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    self.logger.warning(
                        'Can save best model only with {} available, skipping.'.format(self.monitor)
                    )
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            self.logger.info(
                                '\nEpoch {:05d} {} improved from {:0.5f} to {:0.5f},\nsaving model to {}'.format(
                                    epoch + 1, self.monitor, self.best, current, filepath
                                )
                            )
                        self.best = current
                        if self.save_weights_only:
                            self.model.save_weights(str(filepath), overwrite=True)
                        else:
                            self.model.save(str(filepath), overwrite=True)
                    else:
                        if self.verbose > 0:
                            self.logger.info(
                                '\nEpoch {:05d} {} did not improve from {:0.5f}'.format(
                                    epoch + 1, self.monitor, self.best
                                )
                            )
            else:
                if self.verbose > 0:
                    self.logger.debug(
                        '\nEpoch {:05d} saving model to {}'.format(epoch + 1, filepath)
                    )
                if self.save_weights_only:
                    self.model.save_weights(str(filepath), overwrite=True)
                else:
                    self.model.save(str(filepath), overwrite=True)



import numpy as np
from typing import List, Tuple, Optional
from pathlib import Path
from PIL import Image

IMG_FORMATS = ['JPEG', 'PNG']


def load_image(path: Path, target_size=None) -> Image:
    img = Image.open(path)
    f = img.format  # store format after opening as it gets lost after conversion

    if img.mode != 'RGB':
        # convert to RGBA first to avoid warning
        # we ignore alpha channel if available
        img = img.convert('RGBA').convert('RGB')

    if target_size:
        img = img.resize(target_size)

    img.format = f  # reassign format for later validation checks

    return img


def save_image(img, path: Path):
    img.convert('RGB').save(path)


def validate_image(
    file_name: Path, img_formats: List[str] = IMG_FORMATS
) -> Tuple[bool, Optional[Exception]]:
    """
    Checks whether File is valid image file:
        - file exists
        - file is readable
        - file is an image

    Args:
        file_name: Absolute path of file.

     Returns:
        True if file is valid image file.
        False else.
    """

    valid_image = False
    error = None

    try:
        img = load_image(file_name)

        if img.format in img_formats:
            img.load()  # Pillow uses lazy loading, so need to explicitly load

            valid_image = True

    except Exception as e:
        error = e

    return valid_image, error


def resize_image(img: Image, max_size: int, upscale: bool = False) -> Image:
    """Resizes image while keeping aspect ratio.

    The smaller dimension will be resized to max_size, i.e. a 400x500px image with max_size=300
    will be resized to 300x375px.

    Args:
        img: Pillow image object.
        max_size: Maximum width or height of resized image.
        upscale: If True will upscale small images to max_size.

     Returns:
        Pillow image object.
    """
    width, height = img.size

    if (max_size >= min(width, height)) and not upscale:
        return img

    min_dim = min(width, height)
    new_width = int((max_size / min_dim) * width)
    new_height = int((max_size / min_dim) * height)

    return img.resize((new_width, new_height))


def resize_image_mp(data_tuple: Tuple[str, str, str]):
    image_dir, new_image_dir, image_id = data_tuple
    img = load_image(Path(image_dir) / image_id)
    img = resize_image(img, max_size=300, upscale=False)
    save_image(img, Path(new_image_dir) / image_id)


def random_crop(img: np.array, crop_dims: Tuple[int, int]):
    h, w = img.shape[0], img.shape[1]
    ch, cw = crop_dims[0], crop_dims[1]
    assert h >= ch, 'image height is less than crop height'
    assert w >= cw, 'image width is less than crop width'
    x = np.random.randint(0, w - cw + 1)
    y = np.random.randint(0, h - ch + 1)
    return img[y : (y + ch), x : (x + cw), :]



import click
from imageatm.client import commands
from imageatm.client.config import Config


# creates config object that will be passed from cli to subcommands
pass_config = click.make_pass_decorator(Config, ensure=True)


# cli is the group parent, it gets run before any of the subcommands are run
@click.group()
@pass_config
def cli(config: Config):
    pass


@cli.command()
@click.argument('config-file', type=click.Path())
@click.option('--image-dir', type=click.Path(), help='Directory with image files.')
@click.option('--samples-file', type=click.Path(), help='JSON file with samples.')
@click.option(
    '--job-dir',
    type=click.Path(),
    help=('Directory with train, val, and test samples files and class_mapping file.'),
)
@click.option('--provider', help='Cloud provider, currently supported: [aws].')
@click.option('--instance-type', help='Cloud instance_type [aws].')
@click.option('--region', help='Cloud region [aws].')
@click.option('--vpc-id', help='Cloud VPC id [aws].')
@click.option('--bucket', help='Cloud bucket used for persistence [aws].')
@click.option('--tf-dir', help='Directory with Terraform configs [aws].')
@click.option('--train-cloud', is_flag=True, required=False, help='Run training in cloud [aws].')
@click.option('--destroy', is_flag=True, required=False, help='Destroys cloud.')
@click.option('--resize', is_flag=True, required=False, help='Resizes images in dataprep.')
@click.option('--batch-size', type=int, help='Batch size.', required=False)
@click.option(
    '--epochs-train-dense',
    type=int,
    help='Number of epochs train only dense layer.',
    required=False,
)
@click.option(
    '--epochs-train-all', type=int, help='Number of epochs train all layers.', required=False
)
@click.option(
    '--learning-rate-dense', type=float, help='Learning rate dense layers.', required=False
)
@click.option('--learning-rate-all', type=float, help='Learning rate all layers.', required=False)
@click.option(
    '--base-model-name', help='Pretrained CNN to be used for transfer learning.', required=False
)
@click.option(
    '--create-report', is_flag=True, required=False, help='Create evaluation report via jupyter notebook.'
)
@click.option(
    '--kernel-name',  required=False, help='Kernel-name for juypter notebook.'
)
@click.option(
    '--export-html', is_flag=True, required=False, help='Export evaluation report to html.'
)
@click.option(
    '--export-pdf', is_flag=True, required=False, help='Export evaluation report to pdf.'
)
@click.option(
    '--cloud-tag', help='Tag under which all cloud resources are created.', required=False
)
@pass_config
def pipeline(config: Config, **kwargs):
    """Runs all components for which run=True in config file.

    All activated (run=True) components from config file will be run in sequence. Options overwrite the config file.
    The config file is the only way to define pipeline components.

    Args:
        config-file: Central configuration file.
    """
    commands.pipeline(config, **kwargs)


@cli.command()
@click.option('--config-file', type=click.Path(), help='Central configuration file.')
@click.option('--image-dir', type=click.Path(), help='Directory with image files.')
@click.option('--samples-file', type=click.Path(), help='JSON file with samples.')
@click.option(
    '--job-dir',
    type=click.Path(),
    help=('Directory with train, val, and test samples files and class_mapping file.'),
)
@click.option(
    '--resize',
    is_flag=True,
    required=False,
    help='Resizes images and stores them in _resized subfolder.',
)
@pass_config
def dataprep(config: Config, **kwargs):
    """Run data preparation and create job dir.

    Creates a directory (job_dir) with the following files:

        - train_samples.json

        - val_samples.json

        - test_samples.json

        - class_mapping.json
    """
    commands.dataprep(config, **kwargs)


@cli.command()
@click.option('--config-file', type=click.Path(), help='Central configuration file.')
@click.option('--image-dir', type=click.Path(), help='Directory with image files.')
@click.option(
    '--job-dir',
    type=click.Path(),
    help=('Directory with train, val, and test samples files and class_mapping file.'),
)
@click.option('--provider', help='Cloud provider, currently supported: [aws].')
@click.option('--instance-type', help='Cloud instance_type [aws].')
@click.option('--region', help='Cloud region [aws].')
@click.option('--vpc-id', help='Cloud VPC id [aws].')
@click.option('--bucket', help='Cloud bucket used for persistence [aws].')
@click.option('--tf-dir', help='Directory with Terraform configs [aws].')
@click.option('--train-cloud', is_flag=True, required=False, help='Run training in cloud [aws].')
@click.option('--destroy', is_flag=True, required=False, help='Destroys cloud.')
@click.option('--batch-size', type=int, help='Batch size.', required=False)
@click.option(
    '--epochs-train-dense',
    type=int,
    help='Number of epochs train only dense layer.',
    required=False,
)
@click.option(
    '--epochs-train-all', type=int, help='Number of epochs train all layers.', required=False
)
@click.option(
    '--learning-rate-dense', type=float, help='Learning rate dense layers.', required=False
)
@click.option('--learning-rate-all', type=float, help='Learning rate all layers.', required=False)
@click.option(
    '--base-model-name', help='Pretrained CNN to be used for transfer learning.', required=False
)
@click.option(
    '--cloud-tag', help='Tag under which all cloud resources are created.', required=False
)
@pass_config
def train(config: Config, **kwargs):
    """Train a CNN.

    Fine-tunes an ImageNet pre-trained CNN. The number of classes are derived from train_samples.json.
    After each epoch the model will be evaluated on val_samples.json.

    The best model (based on valuation accuracy) will be saved.

    Args:
        image_dir: Directory with image files.
        job_dir: Directory with train_samples, val_samples, and class_mapping.json.

    """
    commands.train(config, **kwargs)


@cli.command()
@click.option('--config-file', type=click.Path(), help='Central configuration file.')
@click.option('--image-dir', type=click.Path(), help='Directory with image files.')
@click.option(
    '--job-dir', type=click.Path(), help=('Directory with test samples files and trained model.')
)
@click.option(
    '--create-report', is_flag=True, required=False, help='Create evaluation report via jupyter notebook.'
)
@click.option(
    '--kernel-name',  required=False, help='Kernel-name for juypter notebook.'
)
@click.option(
    '--export-html', is_flag=True, required=False, help='Export evaluation report to html.'
)
@click.option(
    '--export-pdf', is_flag=True, required=False, help='Export evaluation report to pdf.'
)
@pass_config
def evaluate(config: Config, **kwargs):
    """Evaluate a trained model.

    Evaluation will be performed on test_samples.json.

    Args:
        image_dir: Directory with image files.
        job_dir: Directory with test_samples.json and class_mapping.json.
    """
    commands.evaluate(config, **kwargs)


@cli.command()
@click.option('--config-file', type=click.Path(), help='Central configuration file.')
@click.option(
    '--job-dir', type=click.Path(), help=('Directory with test samples files and trained model.')
)
@click.option('--provider', help='Cloud provider, currently supported: [aws].')
@click.option('--instance-type', help='Cloud instance_type [aws].')
@click.option('--region', help='Cloud region [aws].')
@click.option('--vpc-id', help='Cloud VPC id [aws].')
@click.option('--bucket', help='Cloud bucket used for persistence [aws].')
@click.option('--tf-dir', help='Directory with Terraform configs [aws].')
@click.option('--train-cloud', is_flag=True, required=False, help='Run training in cloud [aws].')
@click.option('--destroy', is_flag=True, required=False, help='Destroys cloud.')
@click.option('--no-destroy', is_flag=True, required=False, help='Keeps cloud.')
@click.option(
    '--cloud-tag', help='Tag under which all cloud resources are created.', required=False
)
@pass_config
def cloud(config: Config, **kwargs):
    """Launch/destroy a cloud compute instance.

    Launch/destroy cloud instances with Terraform based on Terraform files in tf_dir.
    """
    commands.cloud(config, **kwargs)



import re
import sys
import tqdm
import subprocess
import logging
from typing import Callable
from multiprocessing import Pool, cpu_count


def parallelise(function: Callable, data: list) -> list:
    processes = cpu_count()
    pool = Pool(processes=processes)
    results = list(tqdm.tqdm(pool.imap(function, data), total=len(data)))
    pool.close()
    pool.join()
    return results


def run_cmd(cmd: str, logger: logging.Logger, level: str = 'debug', return_output: bool = False):
    # filter out ANSI color and font formatting
    ansi_re = re.compile(r'\x1b\[[0-9;]*m')

    p = subprocess.Popen(
        cmd,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        universal_newlines=True,
        shell=True,
        bufsize=1,
    )

    # stream process stdout to logger if available
    stdout = []
    line = ''
    while True:
        inchar = p.stdout.read(1)
        line += inchar

        if line[-1:] == '\n':
            new_line = re.sub(ansi_re, '', line[:-1])

            if level == 'debug':
                logger.debug(new_line)
            else:
                logger.info(new_line)

            stdout.append(new_line)
            line = ''

        if not inchar:
            sys.stdout.flush()
            break

    output, error = p.communicate()

    if p.returncode != 0 and error:
        logger.error(error, exc_info=True)
        raise Exception('{}'.format(error))
    if return_output:
        return stdout[-1]



from typing import Optional
from pathlib import Path
from imageatm.utils.logger import get_logger
from imageatm.client.config import Config, validate_config, update_config, update_component_configs


def pipeline(
    config: Config,
    config_file: Path,
    job_dir: Optional[Path] = None,
    image_dir: Optional[Path] = None,
    samples_file: Optional[Path] = None,
    provider: Optional[str] = None,
    instance_type: Optional[str] = None,
    region: Optional[str] = None,
    vpc_id: Optional[str] = None,
    bucket: Optional[str] = None,
    tf_dir: Optional[Path] = None,
    train_cloud: Optional[bool] = None,
    destroy: Optional[bool] = None,
    resize: Optional[bool] = None,
    batch_size: Optional[int] = None,
    learning_rate_dense: Optional[float] = None,
    learning_rate_all: Optional[float] = None,
    epochs_train_dense: Optional[int] = None,
    epochs_train_all: Optional[int] = None,
    base_model_name: Optional[str] = None,
    cloud_tag: Optional[str] = None,
    create_report: Optional[bool] = None,
    kernel_name: Optional[str] = None,
    export_html: Optional[bool] = None,
    export_pdf: Optional[bool] = None,
):
    """Runs the entire pipeline based on config file."""
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        samples_file=samples_file,
        provider=provider,
        instance_type=instance_type,
        region=region,
        vpc_id=vpc_id,
        bucket=bucket,
        tf_dir=tf_dir,
        train_cloud=train_cloud,
        destroy=destroy,
        resize=resize,
        batch_size=batch_size,
        learning_rate_dense=learning_rate_dense,
        learning_rate_all=learning_rate_all,
        epochs_train_dense=epochs_train_dense,
        epochs_train_all=epochs_train_all,
        base_model_name=base_model_name,
        cloud_tag=cloud_tag,
        create_report=create_report,
        kernel_name=kernel_name,
        export_html=export_html,
        export_pdf=export_pdf,
    )

    validate_config(config, config.pipeline)

    Path(config.job_dir).resolve().mkdir(parents=True, exist_ok=True)

    logger = get_logger(__name__, config.job_dir)  # type: ignore

    if 'dataprep' in config.pipeline:
        from imageatm.scripts import run_dataprep

        logger.info(
            '\n********************************\n'
            '******* Data preparation *******\n'
            '********************************'
        )

        dp = run_dataprep(**config.dataprep)

        # update image_dir if images were resized
        if config.dataprep.get('resize', False):
            config.image_dir = dp.image_dir  # type: ignore
            config = update_component_configs(config)

    if 'train' in config.pipeline:
        logger.info(
            '\n********************************\n'
            '*********** Training ***********\n'
            '********************************'
        )

        if config.train.get('cloud'):
            from imageatm.scripts import run_training_cloud

            run_training_cloud(**{**config.cloud, **config.train})
        else:
            from imageatm.scripts import run_training

            run_training(**config.train)

    if 'evaluate' in config.pipeline:
        from imageatm.scripts import run_evaluation

        logger.info(
            '\n********************************\n'
            '********** Evaluation **********\n'
            '********************************'
        )

        run_evaluation(**config.evaluate)

    if 'cloud' in config.pipeline:
        from imageatm.scripts import run_cloud

        run_cloud(**config.cloud)


def dataprep(
    config: Config,
    config_file: Optional[Path] = None,
    image_dir: Optional[Path] = None,
    samples_file: Optional[Path] = None,
    job_dir: Optional[Path] = None,
    resize: Optional[bool] = None,
):
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        samples_file=samples_file,
        resize=resize,
    )

    config.dataprep['run'] = True
    validate_config(config, ['dataprep'])

    from imageatm.scripts import run_dataprep

    run_dataprep(**config.dataprep)


def train(
    config,
    config_file: Optional[Path] = None,
    job_dir: Optional[Path] = None,
    image_dir: Optional[Path] = None,
    provider: Optional[str] = None,
    instance_type: Optional[str] = None,
    region: Optional[str] = None,
    vpc_id: Optional[str] = None,
    bucket: Optional[str] = None,
    tf_dir: Optional[Path] = None,
    train_cloud: Optional[bool] = None,
    destroy: Optional[bool] = None,
    batch_size: Optional[int] = None,
    learning_rate_dense: Optional[float] = None,
    learning_rate_all: Optional[float] = None,
    epochs_train_dense: Optional[int] = None,
    epochs_train_all: Optional[int] = None,
    base_model_name: Optional[str] = None,
    cloud_tag: Optional[str] = None,
):
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        provider=provider,
        instance_type=instance_type,
        region=region,
        vpc_id=vpc_id,
        bucket=bucket,
        tf_dir=tf_dir,
        train_cloud=train_cloud,
        destroy=destroy,
        batch_size=batch_size,
        learning_rate_dense=learning_rate_dense,
        learning_rate_all=learning_rate_all,
        epochs_train_dense=epochs_train_dense,
        epochs_train_all=epochs_train_all,
        base_model_name=base_model_name,
        cloud_tag=cloud_tag,
    )

    config.train['run'] = True

    validate_config(config, ['train'])

    if config.train.get('cloud'):
        from imageatm.scripts import run_training_cloud

        run_training_cloud(**{**config.cloud, **config.train})
    else:
        from imageatm.scripts import run_training

        run_training(**config.train)


def evaluate(
    config: Config,
    config_file: Optional[Path] = None,
    image_dir: Optional[Path] = None,
    job_dir: Optional[Path] = None,
    create_report: Optional[bool] = None,
    kernel_name: Optional[str] = None,
    export_html: Optional[bool] = None,
    export_pdf: Optional[bool] = None,
):
    config = update_config(
        config=config,
        config_file=config_file,
        job_dir=job_dir,
        image_dir=image_dir,
        create_report=create_report,
        kernel_name=kernel_name,
        export_html=export_html,
        export_pdf=export_pdf,
    )

    config.evaluate['run'] = True
    validate_config(config, ['evaluate'])

    from imageatm.scripts import run_evaluation

    run_evaluation(**config.evaluate)


def cloud(
    config,
    job_dir: Optional[Path] = None,
    config_file: Optional[Path] = None,
    provider: Optional[str] = None,
    instance_type: Optional[str] = None,
    region: Optional[str] = None,
    vpc_id: Optional[str] = None,
    bucket: Optional[str] = None,
    tf_dir: Optional[Path] = None,
    train_cloud: Optional[bool] = None,
    destroy: Optional[bool] = None,
    no_destroy: Optional[bool] = None,
    cloud_tag: Optional[str] = None,
):
    config = update_config(
        config=config,
        job_dir=job_dir,
        config_file=config_file,
        provider=provider,
        instance_type=instance_type,
        region=region,
        vpc_id=vpc_id,
        bucket=bucket,
        tf_dir=tf_dir,
        train_cloud=train_cloud,
        destroy=destroy,
        no_destroy=no_destroy,
        cloud_tag=cloud_tag,
    )

    config.cloud['run'] = True
    validate_config(config, ['cloud'])

    from imageatm.scripts import run_cloud

    run_cloud(**config.cloud)



from .run_dataprep import run_dataprep
from .run_training import run_training
from .run_training_cloud import run_training_cloud
from .run_evaluation import run_evaluation
from .run_cloud import run_cloud



from pathlib import Path
from typing import Optional, List
from imageatm.utils.io import load_yaml


class Config:
    def __init__(self) -> None:
        # components
        self.dataprep: dict = {}
        self.train: dict = {}
        self.cloud: dict = {}
        self.evaluate: dict = {}

        # central parameters
        self.image_dir: Optional[Path] = None
        self.job_dir: Optional[Path] = None

        self.pipeline: list = []


def update_component_configs(config: Config) -> Config:
    """Populate central parameters to component configs."""
    if config.image_dir:
        config.dataprep['image_dir'] = config.image_dir
        config.train['image_dir'] = config.image_dir
        config.evaluate['image_dir'] = config.image_dir

    if config.job_dir:
        config.dataprep['job_dir'] = config.job_dir
        config.train['job_dir'] = config.job_dir
        config.evaluate['job_dir'] = config.job_dir
        config.cloud['job_dir'] = config.job_dir

    return config


def update_config(
    config: Config,
    config_file: Optional[Path] = None,
    job_dir: Optional[Path] = None,
    image_dir: Optional[Path] = None,
    samples_file: Optional[Path] = None,
    provider: Optional[str] = None,
    instance_type: Optional[str] = None,
    region: Optional[str] = None,
    vpc_id: Optional[str] = None,
    bucket: Optional[str] = None,
    tf_dir: Optional[Path] = None,
    train_cloud: Optional[bool] = None,
    ## TODO: check if destroy == !no_destroy
    destroy: Optional[bool] = None,
    no_destroy: Optional[bool] = None,
    resize: Optional[bool] = None,
    batch_size: Optional[int] = None,
    learning_rate_dense: Optional[float] = None,
    learning_rate_all: Optional[float] = None,
    epochs_train_dense: Optional[int] = None,
    epochs_train_all: Optional[int] = None,
    base_model_name: Optional[str] = None,
    cloud_tag: Optional[str] = None,
    create_report: Optional[bool] = None,
    kernel_name: Optional[str] = None,
    export_html: Optional[bool] = None,
    export_pdf: Optional[bool] = None,
) -> Config:

    # set defaults
    config.train['cloud'] = False
    config.dataprep['resize'] = False
    config.evaluate['report'] = {
        'create': False,
        'kernel_name': 'imageatm',
        'export_html': False,
        'export_pdf': False
    }
    config.pipeline = []

    # load central config file
    if config_file:
        config_yml: dict = load_yaml(config_file)  # type: ignore

        # populate parameters from config file
        # parameters from config file overwrite defaults
        config.image_dir = config_yml.get('image_dir')
        config.job_dir = config_yml.get('job_dir')

        config.dataprep = {**config.dataprep, **config_yml.get('dataprep', {})}
        config.train = {**config.train, **config_yml.get('train', {})}
        config.evaluate = {**config.evaluate, **config_yml.get('evaluate', {})}
        config.cloud = {**config.cloud, **config_yml.get('cloud', {})}

        # set pipeline
        components = ['dataprep', 'train', 'evaluate', 'cloud']
        config.pipeline = [i for i in components if config_yml.get(i, {}).get('run')]

    # set options
    if job_dir is not None:
        config.job_dir = job_dir

    if image_dir is not None:
        config.image_dir = image_dir

    if samples_file is not None:
        config.dataprep['samples_file'] = samples_file

    if provider is not None:
        config.cloud['provider'] = provider

    if instance_type is not None:
        config.cloud['instance_type'] = instance_type

    if region is not None:
        config.cloud['region'] = region

    if vpc_id is not None:
        config.cloud['vpc_id'] = vpc_id

    if bucket is not None:
        config.cloud['bucket'] = bucket

    if tf_dir is not None:
        config.cloud['tf_dir'] = tf_dir

    if train_cloud is True:
        config.train['cloud'] = True

    if destroy is True:
        config.cloud['destroy'] = True

    if no_destroy is True:
        config.cloud['destroy'] = False

    if resize is True:
        config.dataprep['resize'] = True

    if batch_size is not None:
        config.train['batch_size'] = batch_size

    if learning_rate_dense is not None:
        config.train['learning_rate_dense'] = learning_rate_dense

    if learning_rate_all is not None:
        config.train['learning_rate_all'] = learning_rate_all

    if epochs_train_dense is not None:
        config.train['epochs_train_dense'] = epochs_train_dense

    if epochs_train_all is not None:
        config.train['epochs_train_all'] = epochs_train_all

    if base_model_name is not None:
        config.train['base_model_name'] = base_model_name

    if create_report is True:
        config.evaluate['report']['create'] = create_report

    if kernel_name is not None:
        config.evaluate['report']['kernel_name'] = kernel_name

    if kernel_name is not None:
        config.evaluate['report']['export_html'] = export_html

    if kernel_name is not None:
        config.evaluate['report']['export_pdf'] = export_pdf

    if cloud_tag is not None:
        config.cloud['cloud_tag'] = cloud_tag

    config = update_component_configs(config)

    return config


def get_diff(
    name: str, config: dict, required_keys: List[str], optional_keys: List[str]
) -> List[str]:
    allowed_keys = required_keys + optional_keys

    msg = []
    # check that all required keys are in config keys
    diff = list(set(required_keys).difference(config.keys()))
    if diff:
        msg.append('{} config: missing required parameters [{}]\n'.format(name, ', '.join(diff)))

    # check that config keys are in allowed keys
    diff = list(set(config.keys()).difference(allowed_keys))
    if diff:
        msg.append(
            '{} config: [{}] not in allowed parameters [{}]\n'.format(
                name, ', '.join(diff), ', '.join(allowed_keys)
            )
        )

    return msg


def val_dataprep(config: dict) -> List[str]:
    required_keys = ['image_dir', 'job_dir', 'samples_file', 'run']
    optional_keys = ['resize']

    return get_diff('dataprep', config, required_keys, optional_keys)


def val_train(config: dict) -> List[str]:
    required_keys = ['image_dir', 'job_dir', 'cloud', 'run']
    optional_keys = [
        'batch_size',
        'learning_rate_dense',
        'learning_rate_all',
        'epochs_train_dense',
        'epochs_train_all',
        'base_model_name',
    ]

    return get_diff('train', config, required_keys, optional_keys)


def val_evaluate(config: dict) -> List[str]:
    required_keys = ['image_dir', 'job_dir', 'report', 'run']
    optional_keys: list = []
    msg = get_diff('evaluate', config, required_keys, optional_keys)

    required_keys = ['create', 'kernel_name', 'export_pdf', 'export_html']
    optional_keys: list = []
    msg += get_diff('evaluate', config['report'], required_keys, optional_keys)
    return msg


def val_cloud(config: dict) -> List[str]:
    allowed_providers = ['aws']

    assert config.get('provider') is not None, 'Config error: cloud config: missing provider'

    provider = config.get('provider')
    assert (
        provider in allowed_providers
    ), 'Config error: cloud config: {} not in allowed providers [{}]'.format(
        provider, *allowed_providers
    )

    required_keys = {
        'aws': [
            'run',
            'job_dir',
            'tf_dir',
            'region',
            'vpc_id',
            'instance_type',
            'bucket',
            'destroy',
            'provider',
            'cloud_tag',
        ]
    }
    optional_keys: list = []

    return get_diff('cloud', config, required_keys[provider], optional_keys)


def validate_config(config: Config, components: Optional[list]):
    msgs: list = []

    if components:
        for component in components:
            config_component = getattr(config, component)

            if component == 'dataprep':
                msgs += val_dataprep(config_component)

            if component == 'train':
                msgs += val_train(config_component)

            if component == 'evaluate':
                msgs += val_evaluate(config_component)

            if component == 'cloud':
                msgs += val_cloud(config_component)

    assert len(msgs) == 0, '\nConfig error:\n{}'.format(''.join(msgs))



import itertools
import numpy as np
import matplotlib.pyplot as plt
import nbformat
import papermill as pm
from nbconvert import HTMLExporter, PDFExporter
from os.path import dirname

plt.style.use('ggplot')
from typing import List, Union, Tuple, Any
from pathlib import Path
from sklearn.metrics import confusion_matrix, accuracy_score, classification_report
from vis.visualization import visualize_cam
from imageatm.handlers.image_classifier import ImageClassifier
from imageatm.handlers.data_generator import ValDataGenerator
from imageatm.utils.io import load_json
from imageatm.utils.images import load_image
from imageatm.utils.logger import get_logger
from imageatm.utils.tf_keras import use_multiprocessing, load_model
import imageatm.notebooks


BATCH_SIZE = 16
BASE_MODEL_NAME = 'MobileNet'
MAX_N_CLASSES = 20

USE_MULTIPROCESSING, WORKERS = use_multiprocessing()

TYPE_IMAGE_LIST = List[List[Tuple[int, np.array, dict]]]  # used for type hinting


class Evaluation:
    """Calculates performance metrics for trained models.

    Loads the best model (validation accuracy) from *models* directory in job directory.
    All metrics and graphs are based on *test_samples.json* in job directory.
    Plots will only be shown if number of classes 20 or less.

    Attributes:
        image_dir: Path of image directory.
        job_dir: Path to job directory with samples.
        batch_size: Number of images per batch (default 64).
        base_model_name: Name of pretrained CNN (default MobileNet).
    """

    def __init__(
        self,
        image_dir: str,
        job_dir: str,
        batch_size: int = BATCH_SIZE,
        base_model_name: str = BASE_MODEL_NAME,
        **kwargs
    ) -> None:
        """Inits evaluation component.

        Loads the best model from job directory.
        Creates evaluation directory if app was started from commandline.
        """
        self.image_dir = Path(image_dir).resolve()
        self.job_dir = Path(job_dir).resolve()
        self.batch_size = batch_size
        self.base_model_name = base_model_name

        self.logger = get_logger(__name__, self.job_dir)
        self.samples_test: list = load_json(self.job_dir / 'test_samples.json')  # type: ignore
        self.class_mapping: dict = load_json(self.job_dir / 'class_mapping.json')  # type: ignore
        self.n_classes = len(self.class_mapping)
        self.classes = [str(self.class_mapping[str(i)]) for i in range(self.n_classes)]
        self.y_true = np.array([i['label'] for i in self.samples_test])
        self.figures = []

        self._determine_plot_params()
        self._load_best_model()
        self._create_evaluation_dir()

    def _determine_plot_params(self):
        """Determines fontsizes and checks whether ipython kernel is present.

        Plots will only be shown if in ipython, otherwise saved as files.
        """
        self.fontsize_title = 18 if self.n_classes < 4 else 18
        self.fontsize_label = 14 if self.n_classes < 4 else 14
        self.fontsize_ticks = 12 if self.n_classes < 4 else 12
        self.mode_ipython = True if self._is_in_ipython_mode() else False

    def _is_in_ipython_mode(self):
        try:
            __IPYTHON__
            return True

        except NameError:
            ## TODO: Is this obsolete? Please remove!
            # Suppress figure window in terminal
            # https://matplotlib.org/faq/howto_faq.html#generate-images-without-having-a-window-appear
            import matplotlib
            matplotlib.use('Agg')

            return False

    def _load_best_model(self):
        """Loads best performing model from job_dir."""
        self.logger.info('\n****** Load model ******\n')
        best_model_file = self._determine_best_modelfile()
        self.best_model_file = Path(best_model_file).resolve()
        self.best_model = load_model(self.best_model_file)

        self.logger.info('loaded {}'.format(self.best_model_file))

    def _determine_best_modelfile(self):
        """Determines best performing model from job_dir."""
        job_path = self.job_dir / 'models'
        model_files = list(job_path.glob('**/*.hdf5'))
        max_acc_idx = np.argmax([m.name.split('_')[3][:5] for m in model_files])
        return model_files[max_acc_idx]

    def _create_evaluation_dir(self):
        """Creates evaluation dir for reporting."""
        if not self.mode_ipython:
            evaluation_dir_name = self.best_model_file.name.split('.hdf5')[0]
            self.evaluation_dir = self.job_dir / 'evaluation_{}'.format(evaluation_dir_name)

            self.evaluation_dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _get_probabilities_prediction(predictions_dist: List[List[float]]) -> List[float]:
        index = np.argmax(predictions_dist, axis=1)
        prob = [pred[index] for pred, index in zip(predictions_dist, index)]
        return prob

    def _make_prediction_on_test_set(self):
        """Makes prediction on test set."""

        self.classifier = ImageClassifier(
            base_model_name=self.base_model_name,
            n_classes=self.n_classes,
            weights=None,
            dropout_rate=None,
            learning_rate=None,
            loss=None,
        )
        self.classifier.model = self.best_model

        self.data_generator = ValDataGenerator(
            samples=self.samples_test,
            image_dir=self.image_dir,
            batch_size=self.batch_size,
            n_classes=self.n_classes,
            basenet_preprocess=self.classifier.get_preprocess_input(),
        )

        predictions_dist = self.classifier.predict_generator(
            data_generator=self.data_generator,
            workers=WORKERS,
            use_multiprocessing=USE_MULTIPROCESSING,
            verbose=1,
        )

        self.y_pred = np.argmax(predictions_dist, axis=1)
        self.y_pred_prob = self._get_probabilities_prediction(predictions_dist=predictions_dist)

    def _plot_test_set_distribution(self, figsize: (float, float) = [8, 5]):
        """Plots bars with number of samples for each label in test set."""
        assert self.mode_ipython, 'Plotting is only possible when in ipython-mode'

        if self.n_classes > MAX_N_CLASSES:
            self.logger.info('\nPlotting only for max {} classes\n'.format(MAX_N_CLASSES))
            return

        x_tick_marks = np.arange(self.n_classes)
        y_values = np.bincount(self.y_true)
        title = 'Number of images in test set: {}'.format(len(self.samples_test))

        fig = plt.figure(figsize=figsize)
        plt.rcParams["axes.grid"] = True
        plt.bar(x_tick_marks, y_values)
        plt.title(title, fontsize=self.fontsize_title)
        plt.xlabel('Label', fontsize=self.fontsize_label)
        plt.ylabel('Number of images', fontsize=self.fontsize_label)
        plt.xticks(x_tick_marks, self.classes, fontsize=self.fontsize_ticks, rotation=30)

        plt.tight_layout()
        plt.show()

    def _print_test_set_distribution(self):
        """Prints distribution for labels in test set."""
        assert not self.mode_ipython, 'Printing is recommended when not in ipython-mode'

        max_length = len(max(self.classes, key=len))
        y_values = np.bincount(self.y_true)
        for i, c in enumerate(self.classes):
            label = c + ' ' * (max_length - len(c))
            self.logger.info("{}\t{}". format(label, y_values[i]))

    def _print_classification_report(self):
        """Prints classification for labels in test set."""
        cr = classification_report(
            y_true=self.y_true, y_pred=self.y_pred, target_names=self.classes, output_dict=True
        )

        metrics = ['precision', 'recall', 'f1-score', 'support']
        categories = self.classes.copy()
        categories.extend(['macro avg', 'weighted avg'])

        max_length = len(max(self.classes, key=len))
        self.logger.info("{}\t{}\t{}\t{}\t{}". format(' ' * max_length, 'prec', "rec", "f1", "support"))
        for c in categories:
            label = c + ' ' * (max_length - len(c))
            line_output = "{}\t".format(label)
            for m in metrics:
                if m == 'support':
                    line_output += "{}\t".format(cr[c][m])
                else:
                    line_output += "{0:.2f}\t".format(cr[c][m])
            self.logger.info(line_output)

    def _plot_confusion_matrix(self, figsize: (float, float) = [9, 9], precision: bool = False):
        """Plots normalized confusion matrix."""
        assert self.mode_ipython, 'Plotting is only possible when in ipython-mode'

        if self.n_classes > MAX_N_CLASSES:
            self.logger.info('\nPlotting only for max {} classes\n'.format(MAX_N_CLASSES))
            return

        (title, xlabel, ylabel) = \
            ('Confusion matrix (precision)', 'True label', 'Predicted label') if precision \
            else ('Confusion matrix (recall)', 'Predicted label', 'True label')

        cm = confusion_matrix(y_true=self.y_pred, y_pred=self.y_true) if precision \
            else confusion_matrix(y_true=self.y_true, y_pred=self.y_pred)
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        tick_marks = np.arange(self.n_classes)

        fig = plt.figure(figsize=figsize)
        plt.rcParams["axes.grid"] = False
        plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues, vmin=0, vmax=1)
        plt.colorbar()
        plt.title(title, fontsize=self.fontsize_title)
        plt.xlabel(xlabel, fontsize=self.fontsize_label)
        plt.ylabel(ylabel, fontsize=self.fontsize_label)
        plt.xticks(tick_marks, self.classes, rotation=45, fontsize=self.fontsize_ticks, ha='right')
        plt.yticks(tick_marks, self.classes, fontsize=self.fontsize_ticks)

        thresh = cm.max() / 2.0
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            plt.text(
                j,
                i,
                '{:.2f}'.format(cm[i, j]),
                horizontalalignment='center',
                color='white' if cm[i, j] > thresh else 'black',
                fontsize=self.fontsize_ticks,
            )

        plt.tight_layout()
        plt.show()

    def _print_confusion_matrix(self, precision: bool = False):
        """Prints normalized confusion matrix."""
        assert not self.mode_ipython, 'Printing is recommended when not in ipython-mode'

        cm = confusion_matrix(y_true=self.y_pred, y_pred=self.y_true) if precision \
            else confusion_matrix(y_true=self.y_true, y_pred=self.y_pred)
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

        max_length = len(max(self.classes, key=len))
        for i, c in enumerate(self.classes):
            label = c + ' ' * (max_length - len(c))
            line_output = "{}\t".format(label)
            for x in cm[i].tolist():
                line_output += "{0:.2f}\t".format(x)
            self.logger.info(line_output)

    def _plot_correct_wrong_examples(self):
        """Plots correct and wrong examples for each label in test set."""
        assert self.mode_ipython, 'Plotting is only possible when in ipython-mode'

        if self.n_classes > MAX_N_CLASSES:
            self.logger.info('\nPlotting only for max {} classes\n'.format(MAX_N_CLASSES))
            return

        for i in range(len(self.classes)):
            c, w = self.get_correct_wrong_examples(label=i)
            self.visualize_images(c, title='Label: "{}" (correct predicted)'.format(self.classes[i]), show_heatmap=True, n_plot=3)
            self.visualize_images(w, title='Label: "{}" (wrong predicted)'.format(self.classes[i]), show_heatmap=True, n_plot=3)

    def _create_report(self, report_kernel_name:str, report_export_html:bool, report_export_pdf:bool):
        """Creates report from notebook-template and stores it in different formats all figures.

            - Jupyter Notebook
            - HTML
            - PDF
        """
        assert not self.mode_ipython, 'Create report is only possible when not in ipython mode'

        filepath_template = dirname(imageatm.notebooks.__file__) + '/evaluation_template.ipynb'
        filepath_notebook = self.evaluation_dir / 'evaluation_report.ipynb'
        filepath_html = self.evaluation_dir / 'evaluation_report.html'
        filepath_pdf = self.evaluation_dir / 'evaluation_report.pdf'

        pm.execute_notebook(
            str(filepath_template),
            str(filepath_notebook),
            parameters=dict(
                image_dir=str(self.image_dir),
                job_dir=str(self.job_dir)
            ),
            kernel_name=report_kernel_name
        )

        with open(filepath_notebook) as f:
            nb = nbformat.read(f, as_version=4)

        if report_export_html:
            self.logger.info('\n****** Create HTML ******\n')
            with open(filepath_notebook) as f:
                nb = nbformat.read(f, as_version=4)

            html_exporter = HTMLExporter()
            html_data, resources = html_exporter.from_notebook_node(nb)

            with open(filepath_html, 'w') as f:
                f.write(html_data)
                f.close()

        if report_export_pdf:
            self.logger.info('\n****** Create PDF ******\n')

            pdf_exporter = PDFExporter()
            pdf_exporter.template_file = dirname(imageatm.notebooks.__file__) + '/tex_templates/evaluation_report.tplx'
            pdf_data, resources = pdf_exporter.from_notebook_node(nb, resources={
                'metadata': {'name': 'Evaluation Report'}
            })

            with open(filepath_pdf, 'wb') as f:
                f.write(pdf_data)
                f.close()

    # TO-DO: Enforce string or integer but not both at the same time
    def get_correct_wrong_examples(
        self, label: Union[int, str]
    ) -> Tuple[TYPE_IMAGE_LIST, TYPE_IMAGE_LIST]:
        """Gets correctly and wrongly predicted samples for a given label.

        Args:
            label: int or str (label for which the predictions should be considered).

        Returns:
            (correct, wrong): Tuple of two image lists.
        """
        correct = []
        wrong = []

        if type(label) == str:
            class_mapping_inv = {v: k for k, v in self.class_mapping.items()}
            label = int(class_mapping_inv[label])

        for i, sample in enumerate(self.samples_test):
            if self.y_true[i] == label:
                image_file = self.image_dir / sample['image_id']
                if self.y_true[i] == self.y_pred[i]:
                    correct.append([i, load_image(image_file, target_size=(224, 224)), sample])
                else:
                    wrong.append([i, load_image(image_file, target_size=(224, 224)), sample])

        return correct, wrong

    def visualize_images(
        self, image_list: TYPE_IMAGE_LIST, title: str = 'Images for visualisation', show_heatmap: bool = False, n_plot: int = 20
    ):
        """Visualizes images in a sample list.

        Args:
            image_list: sample list.
            show_heatmap: boolean (generates a gradient based class activation map (grad-CAM), default False).
            n_plot: maximum number of plots to be shown (default 20).
        """
        assert self.mode_ipython, 'Plotting is only possible when in ipython-mode'

        if len(image_list) == 0:
            print('Empty list.')
            return
        else:
            n_rows = min(n_plot, len(image_list))
            n_cols = 2 if show_heatmap else 1

            figsize = [5 * n_cols, 4 * n_rows]
            fig = plt.figure(figsize=figsize)
            fig.suptitle(title, fontsize=self.fontsize_title)

            plot_count = 1
            for (i, img, sample) in image_list[:n_rows]:
                plt.subplot(n_rows, n_cols, plot_count)
                plt.imshow(img)
                plt.axis('off')
                plt.title(
                    'true: {}, predicted: {} ({})'.format(
                        self.class_mapping[str(self.y_true[i])],
                        self.class_mapping[str(self.y_pred[i])],
                        str(round(self.y_pred_prob[i], 2)),
                    )
                )
                plot_count += 1

                if show_heatmap is True:
                    heatmap = visualize_cam(
                        model=self.classifier.model,
                        layer_idx=89,
                        filter_indices=[self.y_pred[i]],
                        seed_input=self.classifier.get_preprocess_input()(
                            np.array(img).astype(np.float32)
                        ),
                    )
                    plt.subplot(n_rows, n_cols, plot_count)
                    plt.imshow(img)
                    plt.imshow(heatmap, alpha=0.7)
                    plt.axis('off')
                    plot_count += 1

            plt.show()

    def run(self,
            report_create: bool = False,
            report_kernel_name: str = 'imageatm',
            report_export_html: bool = False,
            report_export_pdf: bool = False
        ):
        """Runs evaluation pipeline on the best model found in job directory for the specific test set:

            - Makes prediction on test set
            - Plots test set distribution
            - Plots classification report (accuracy, precision, recall)
            - Plots confusion matrix (on precsion and on recall)
            - Plots correct and wrong examples

           If not in ipython mode an evaluation report is created.

        Args:
        	report_create: boolean (create ipython kernel)
            report_kernel_name: str (name of ipython kernel)
            report_export_html: boolean (exports report to html).
            report_export_pdf: boolean (exports report to pdf).
        """
        if self.mode_ipython:
            self.logger.info('\n****** Make prediction on test set ******\n')
            self._make_prediction_on_test_set()

            self.logger.info('\n****** Plot distribution on test set ******\n')
            self._plot_test_set_distribution(figsize=[8, 5])

            self.logger.info('\n****** Plot classification report ******\n')
            # self._plot_classification_report(figsize=[4 + self.n_classes*0.5, 4 + self.n_classes*0.5])
            self._print_classification_report()

            self.logger.info('\n****** Plot confusion matrix (recall) ******\n')
            self._plot_confusion_matrix(figsize=[4 + self.n_classes*0.5, 4 + self.n_classes*0.5])

            self.logger.info('\n****** Plot confusion matrix (precision) ******\n')
            self._plot_confusion_matrix(figsize=[4 + self.n_classes*0.5, 4 + self.n_classes*0.5], precision=True)

            self.logger.info('\n****** Plot correct and wrong examples ******\n')
            self._plot_correct_wrong_examples()

        elif report_create:
            self.logger.info('\n****** Create Jupyter Notebook (this may take a while) ******\n')
            self._create_report(report_kernel_name, report_export_html, report_export_pdf)

        else:
            self.logger.info('\n****** Make prediction on test set ******\n')
            self._make_prediction_on_test_set()

            self.logger.info('\n****** Print distribution on test set ******\n')
            self._print_test_set_distribution()

            self.logger.info('\n****** Print classification report ******\n')
            self._print_classification_report()

            self.logger.info('\n****** Print confusion matrix (recall) ******\n')
            self._print_confusion_matrix()

            self.logger.info('\n****** Print confusion matrix (precision) ******\n')
            self._print_confusion_matrix(precision=True)



import numpy as np
from typing import List, Callable, Tuple
from pathlib import Path
from imageatm.utils.images import load_image, random_crop
from keras.utils import Sequence, to_categorical


class DataGenerator(Sequence):
    """Class inherits from Keras Sequence base object, allows to use multiprocessing in .fit_generator.

    DataGenerator is extended by these classes:
        - TrainDataGenerator
        - ValDataGenerator.

    Attributes:
        samples: Dictionary of samples to generate data from.
        image_dir: Path of image directory.
        batch_size: Number of images per batch.
        n_classes: Number of classes in dataset.
        basenet_preprocess: Basenet specific preprocessing function.
        img_load_dims: Dimensions that images get resized into when loaded.
        train: If set to True samples are shuffled before each epoch and images are cropped once.
    """

    def __init__(
        self,
        samples: List[dict],
        image_dir: str,
        batch_size: int,
        n_classes: int,
        basenet_preprocess: Callable,
        img_load_dims: Tuple[int, int],
        train: bool,
    ) -> None:
        """Inits DataGenerator object.

        If *train* set *True* then samples are shuffled on init.
        """
        self.samples = samples
        self.image_dir = Path(image_dir)
        self.batch_size = batch_size
        self.n_classes = n_classes
        self.basenet_preprocess = basenet_preprocess
        self.img_load_dims = img_load_dims
        self.train = train
        self.on_epoch_end()

    def on_epoch_end(self):
        """Method called at the end of every epoch.

        If *train* set *True* then samples are shuffled.
        """
        self.indexes = np.arange(len(self.samples))
        if self.train is True:
            np.random.shuffle(self.indexes)

    def __len__(self):
        """Number of batches in the Sequence."""
        return int(np.ceil(len(self.samples) / self.batch_size))

    def __getitem__(self, index: int) -> Tuple[np.array, np.array]:
        """Gets batch at position `index`.

        If *train* set *True* then images will be cropped by *img_crop_dims*.
        """
        batch_indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]
        batch_samples = [self.samples[i] for i in batch_indexes]
        X, y = self._data_generator(batch_samples)
        return X, y

    def _data_generator(self, batch_samples: List[dict]) -> Tuple[np.array, np.array]:
        """Generates data from samples in specified batch."""
        #  initialize images and labels tensors for faster processing
        dims = self.img_crop_dims if self.train == True else self.img_load_dims
        X = np.empty((len(batch_samples), *dims, 3))
        y = np.empty((len(batch_samples), self.n_classes))

        for i, sample in enumerate(batch_samples):
            # load and randomly augment image
            img_file = self.image_dir / sample['image_id']
            img = np.asarray(load_image(img_file, self.img_load_dims))
            if self.train == True:
                img = random_crop(img, self.img_crop_dims)
            X[i,] = img

            # TODO: more efficient by preprocessing
            y[i,] = to_categorical([sample['label']], num_classes=self.n_classes)

        # apply basenet specific preprocessing
        # input is 4D numpy array of RGB values within [0, 255]
        X = self.basenet_preprocess(X)

        return X, y


class TrainDataGenerator(DataGenerator):
    """Class inherits from DataGenerator.

    Per default images will be cropped and samples are shuffled before each epoch.

    Attributes:
        samples: Dictionary of samples to generate data from.
        image_dir: Path of image directory.
        batch_size: Number of images per batch.
        n_classes: Number of classes in dataset.
        basenet_preprocess: Basenet specific preprocessing function.
        img_load_dims: Dimensions that images get resized into when loaded (default (256, 256)).
        img_crop_dims: Dimensions that images get resized into when loaded (default (224, 224)).
        train: If set to True samples are shuffled before each epoch and images are cropped once (default True).
    """

    def __init__(
        self,
        samples: List[dict],
        image_dir: str,
        batch_size: int,
        n_classes: int,
        basenet_preprocess: Callable,
        img_load_dims: Tuple[int, int] = (256, 256),
        img_crop_dims: Tuple[int, int] = (224, 224),
        train: bool = True,
    ) -> None:
        """Inits TrainDataGenerator object.

        Per default samples are shuffled on init."""
        super(TrainDataGenerator, self).__init__(
            samples, image_dir, batch_size, n_classes, basenet_preprocess, img_load_dims, train
        )
        self.img_crop_dims = img_crop_dims  # dimensions that images get randomly cropped to


class ValDataGenerator(DataGenerator):
    """Class inherits from DataGenerator.

    Per default neither images are cropped nor samples are shuffled.

    Attributes:
        samples: Dictionary of samples to generate data from.
        image_dir: Path of image directory.
        batch_size: Number of images per batch.
        n_classes: Number of classes in dataset.
        basenet_preprocess: Basenet specific preprocessing function.
        img_load_dims: Dimensions that images get resized into when loaded (default (224, 224)).
        train: If set to True samples are shuffled before each epoch and images are cropped once (default False).
    """

    def __init__(
        self,
        samples: List[dict],
        image_dir: str,
        batch_size: int,
        n_classes: int,
        basenet_preprocess: Callable,
        img_load_dims: Tuple[int, int] = (224, 224),
        train: bool = False,
    ) -> None:
        """Inits TrainDataGenerator object."""
        super(ValDataGenerator, self).__init__(
            samples, image_dir, batch_size, n_classes, basenet_preprocess, img_load_dims, train
        )



import typing
import tensorflow as tf
from keras import backend as K
from keras.callbacks import ReduceLROnPlateau, EarlyStopping
from pathlib import Path
from imageatm.utils.io import load_json
from imageatm.utils.logger import get_logger
from imageatm.utils.tf_keras import use_multiprocessing, LoggingMetrics, LoggingModels
from imageatm.handlers.data_generator import TrainDataGenerator, ValDataGenerator
from imageatm.handlers.image_classifier import ImageClassifier


tf.logging.set_verbosity(tf.logging.ERROR)

BATCH_SIZE = 64
DROPOUT_RATE = 0.75
BASE_MODEL_NAME = 'MobileNet'
LOSS = 'categorical_crossentropy'
EPOCHS_TRAIN_DENSE = 100
EPOCHS_TRAIN_ALL = 100
LEARNING_RATE_DENSE = 0.001
LEARNING_RATE_ALL = 0.0003


class Training:
    """Builds model and runs training.

    The following pretrained CNNs from Keras can be used for transfer learning:

    - Xception
    - VGG16
    - VGG19
    - ResNet50, ResNet101, ResNet152
    - ResNet50V2, ResNet101V2, ResNet152V2
    - ResNeXt50, ResNeXt101
    - InceptionV3
    - InceptionResNetV2
    - MobileNet
    - MobileNetV2
    - DenseNet121, DenseNet169, DenseNet201
    - NASNetLarge, NASNetMobile

    Training is split into two phases, at first only the last dense layer gets
    trained, and then all layers are trained. The maximum number of epochs for
    each phase is set by *epochs_train_dense* (default: 100) and
    *epochs_train_all* (default: 100), respectively. Similarly,
    *learning_rate_dense* (default: 0.001) and *learning_rate_all*
    (default: 0.0003) can be set.

    For each phase the learning rate is reduced after a patience period if no
    improvement in validation accuracy has been observed. The patience period
    depends on the average number of samples per class:

    - if n_per_class < 200: patience = 5 epochs
    - if n_per_class >= 200 and < 500: patience = 4 epochs
    - if n_per_class >= 500: patience = 2 epochs

    The training is stopped early after a patience period that is three times
    the learning rate patience to allow for two learning rate adjustments
    with no validation accuracy improvement before stopping training.

    Attributes:
        image_dir: Directory with images used for training.
        job_dir: Directory with train_samples.json, val_samples.json,
                 and class_mapping.json.
        epochs_train_dense: Maximum number of epochs to train dense layers (default 100).
        epochs_train_all: Maximum number of epochs to train all layers (default 100).
        learning_rate_dense: Learning rate for dense training phase (default 0.001).
        learning_rate_all: Learning rate for all training phase (default 0.0003).
        batch_size: Number of images per batch (default 64).
        dropout_rate: Fraction of nodes before output layer set to random value (default 0.75).
        base_model_name: Name of pretrained CNN (default MobileNet).
    """

    def __init__(
        self,
        image_dir: str,
        job_dir: str,
        epochs_train_dense: typing.Union[int, str] = EPOCHS_TRAIN_DENSE,
        epochs_train_all: typing.Union[int, str] = EPOCHS_TRAIN_ALL,
        learning_rate_dense: typing.Union[float, str] = LEARNING_RATE_DENSE,
        learning_rate_all: typing.Union[float, str] = LEARNING_RATE_ALL,
        batch_size: typing.Union[int, str] = BATCH_SIZE,
        dropout_rate: typing.Union[float, str] = DROPOUT_RATE,
        base_model_name: str = BASE_MODEL_NAME,
        loss: str = LOSS,
        **kwargs,
    ) -> None:
        """Inits training component.

        Checks whether multiprocessing is available and sets number of workers for training.
        """
        self.image_dir = Path(image_dir).resolve()
        self.job_dir = Path(job_dir).resolve()

        self.logger = get_logger(__name__, self.job_dir)
        self.samples_train = load_json(self.job_dir / 'train_samples.json')
        self.samples_val = load_json(self.job_dir / 'val_samples.json')
        self.class_mapping = load_json(self.job_dir / 'class_mapping.json')
        self.n_classes = len(self.class_mapping)

        self.epochs_train_dense = int(epochs_train_dense)
        self.epochs_train_all = int(epochs_train_all)
        self.learning_rate_dense = float(learning_rate_dense)
        self.learning_rate_all = float(learning_rate_all)
        self.batch_size = int(batch_size)
        self.dropout_rate = float(dropout_rate)
        self.base_model_name = base_model_name
        self.loss = loss
        self.use_multiprocessing, self.workers = use_multiprocessing()

    def _set_patience(self):
        """Adjust patience for early stopping and learning rate schedule
        based on training set size.
        """
        n_per_class = int(len(self.samples_train) / self.n_classes)

        self.patience_learning_rate = 5
        if n_per_class >= 200:
            self.patience_learning_rate = 4

        if n_per_class >= 500:
            self.patience_learning_rate = 2

        self.patience_early_stopping = 3 * self.patience_learning_rate

        self.logger.info('Early stopping patience: {}'.format(self.patience_early_stopping))
        self.logger.info('Learning rate patience: {}'.format(self.patience_learning_rate))

    def _build_model(self):
        self.classifier = ImageClassifier(
            self.base_model_name,
            self.n_classes,
            self.learning_rate_dense,
            self.dropout_rate,
            self.loss,
        )
        self.classifier.build()

    def _fit_model(self):
        training_generator = TrainDataGenerator(
            self.samples_train,
            self.image_dir,
            self.batch_size,
            self.n_classes,
            self.classifier.get_preprocess_input(),
        )

        validation_generator = ValDataGenerator(
            self.samples_val,
            self.image_dir,
            self.batch_size,
            self.n_classes,
            self.classifier.get_preprocess_input(),
        )

        # TODO: initialize callbacks TensorBoardBatch
        # tensorboard = TensorBoardBatch(log_dir=Path(job_dir).resolve() / 'logs')

        model_save_name = (
            'model_' + self.base_model_name.lower() + '_{epoch:02d}_{val_acc:.3f}.hdf5'
        )
        model_dir = self.job_dir / 'models'
        model_dir.mkdir(parents=True, exist_ok=True)

        logging_metrics = LoggingMetrics(logger=self.logger)
        logging_models = LoggingModels(
            logger=self.logger,
            filepath=str(model_dir / model_save_name),
            monitor='val_acc',
            verbose=1,
            save_best_only=True,
            save_weights_only=False,
        )

        def _train_dense_layers():
            if self.epochs_train_dense > 0:
                self.logger.info('\n****** Train dense layers ******\n')

                min_lr = self.learning_rate_dense / 10
                reduce_lr = ReduceLROnPlateau(
                    monitor='val_acc',
                    factor=0.3162,
                    patience=self.patience_learning_rate,
                    min_lr=min_lr,
                    verbose=1,
                )

                early_stopping = EarlyStopping(
                    monitor='val_acc',
                    min_delta=0.002,
                    patience=self.patience_early_stopping,
                    verbose=1,
                    mode='auto',
                    baseline=None,
                    restore_best_weights=True,
                )

                # freeze convolutional layers in base net
                for layer in self.classifier.get_base_layers():
                    layer.trainable = False

                self.classifier.compile()
                # self.classifier.summary()

                self.hist_dense = self.classifier.fit_generator(
                    generator=training_generator,
                    validation_data=validation_generator,
                    epochs=self.epochs_train_dense,
                    verbose=1,
                    use_multiprocessing=self.use_multiprocessing,
                    workers=self.workers,
                    max_queue_size=30,
                    callbacks=[logging_metrics, logging_models, reduce_lr, early_stopping],
                )

        def _train_all_layers():
            if self.epochs_train_all > 0:
                self.logger.info('\n****** Train all layers ******\n')

                min_lr = self.learning_rate_all / 10
                reduce_lr = ReduceLROnPlateau(
                    monitor='val_acc',
                    factor=0.3162,
                    patience=self.patience_learning_rate,
                    min_lr=min_lr,
                    verbose=1,
                )

                early_stopping = EarlyStopping(
                    monitor='val_acc',
                    min_delta=0.002,
                    patience=self.patience_early_stopping,
                    verbose=1,
                    mode='auto',
                    baseline=None,
                    restore_best_weights=False,
                )

                # unfreeze all layers
                for layer in self.classifier.get_base_layers():
                    layer.trainable = True

                self.classifier.set_learning_rate(self.learning_rate_all)

                self.classifier.compile()
                # self.classifier.summary()

                self.hist_all = self.classifier.fit_generator(
                    generator=training_generator,
                    validation_data=validation_generator,
                    epochs=self.epochs_train_dense + self.epochs_train_all,
                    initial_epoch=self.epochs_train_dense,
                    verbose=1,
                    use_multiprocessing=self.use_multiprocessing,
                    workers=self.workers,
                    max_queue_size=30,
                    callbacks=[logging_metrics, logging_models, reduce_lr, early_stopping],
                )

        self._set_patience()
        _train_dense_layers()
        _train_all_layers()

        K.clear_session()

    def run(self):
        """Builds the model and runs training.
        """
        self._build_model()
        self._fit_model()



from imageatm.components import Evaluation


def run_evaluation(
	image_dir: str,
	job_dir: str,
	report: object,
	**kwargs,
):
    eval = Evaluation(image_dir=image_dir, job_dir=job_dir, **kwargs)
    eval.run(
		report_create = report['create'],
		report_kernel_name = report['kernel_name'],
		report_export_html = report['export_html'],
		report_export_pdf = report['export_pdf']
	)



from imageatm.components import AWS


def run_training_cloud(
    image_dir: str,
    job_dir: str,
    provider: str,
    tf_dir: str,
    region: str,
    instance_type: str,
    vpc_id: str,
    bucket: str,
    destroy: bool,
    cloud_tag: str,
    **kwargs
):
    if provider == 'aws':
        cloud = AWS(
            tf_dir=tf_dir,
            region=region,
            instance_type=instance_type,
            vpc_id=vpc_id,
            s3_bucket=bucket,
            job_dir=job_dir,
            cloud_tag=cloud_tag,
        )

    cloud.init()
    cloud.apply()
    cloud.train(image_dir=image_dir, job_dir=job_dir, **kwargs)

    if destroy:
        cloud.destroy()



import importlib
from typing import Callable
from keras.models import Model
from keras.layers import Dropout, Dense
from keras.optimizers import Adam
from keras.callbacks import History
from imageatm.handlers.data_generator import DataGenerator


class ImageClassifier:
    """Class that represents the classifier for images.

    The following pretrained CNNs from Keras can be used as base model:
    - Xception
    - VGG16
    - VGG19
    - ResNet50, ResNet101, ResNet152
    - ResNet50V2, ResNet101V2, ResNet152V2
    - ResNeXt50, ResNeXt101
    - InceptionV3
    - InceptionResNetV2
    - MobileNet
    - MobileNetV2
    - DenseNet121, DenseNet169, DenseNet201
    - NASNetLarge, NASNetMobile

    Attributes:
        base_model_name: Name of Keras base model.
        n_classes: Number of classes.
        learning_rate: Learning rate for training phase.
        dropout_rate: Fraction set randomly.
        loss: A loss function as one of two parameters to compile the model.
        weights: Pretrained weights the model architecture is loaded with (default imagenet).
    """

    def __init__(
        self,
        base_model_name: str,
        n_classes: int,
        learning_rate: float,
        dropout_rate: float,
        loss: str,
        weights: str = 'imagenet',
    ) -> None:
        """Inits ImageClassifier object.

        Loads Keras base_module specified by base_model_name.
        """
        self.n_classes = n_classes
        self.base_model_name = base_model_name
        self.learning_rate = learning_rate
        self.dropout_rate = dropout_rate
        self.loss = loss
        self.weights = weights
        self._load_base_module()

    def _load_base_module(self):
        """Loads Keras base_module specified by base_model_name. """
        if self.base_model_name == 'InceptionV3':
            self.base_module = importlib.import_module('keras.applications.inception_v3')

        elif self.base_model_name == 'InceptionResNetV2':
            self.base_module = importlib.import_module('keras.applications.inception_resnet_v2')

        elif self.base_model_name in ['NASNetLarge', 'NASNetMobile']:
            self.base_module = importlib.import_module('keras.applications.nasnet')

        elif self.base_model_name in ['DenseNet121', 'DenseNet169', 'DenseNet201']:
            self.base_module = importlib.import_module('keras.applications.densenet')

        elif self.base_model_name in ['ResNet50', 'ResNet101', 'ResNet152']:
            self.base_module = importlib.import_module('keras.applications.resnet')

        elif self.base_model_name in ['ResNet50V2', 'ResNet101V2', 'ResNet152V2']:
            self.base_module = importlib.import_module('keras.applications.resnet_v2')

        elif self.base_model_name in ['ResNeXt50', 'ResNeXt101']:
            self.base_module = importlib.import_module('keras.applications.resnext')

        else:
            self.base_module = importlib.import_module(
                'keras.applications.' + self.base_model_name.lower()
            )

    def get_base_layers(self) -> list:
        """ Gets layers of classifiers' base model

        Returns:
            base_layers: list of layers
        """
        return self.base_model.layers

    def get_preprocess_input(self) -> Callable:
        """ Gets preprocess_input of classifiers' base_module

        Returns:
            preprocess_input: Callable
        """
        return self.base_module.preprocess_input

    def set_learning_rate(self, learning_rate: float):
        """
        sets classifiers' learning_rate

        Args:
            learning_rate: the learning_rate.
        """
        self.learning_rate = learning_rate

    def build(self):
        """
        Builds classifiers' model.

        The following steps will be performed in sequence:
            - Loads a pretrained base model.
            - Adds dropout and dense layer to the model.
            - Sets classifiers' model.
        """
        BaseCnn = getattr(self.base_module, self.base_model_name)
        self.base_model = BaseCnn(
            input_shape=(224, 224, 3), weights=self.weights, include_top=False, pooling='avg'
        )

        x = Dropout(self.dropout_rate)(self.base_model.output)
        x = Dense(units=self.n_classes, activation='softmax')(x)

        self.model = Model(self.base_model.inputs, x)

    def compile(self):
        """ Configures classifiers' model for training."""
        self.model.compile(
            optimizer=Adam(lr=self.learning_rate), loss=self.loss, metrics=['accuracy']
        )

    def fit_generator(self, **kwargs) -> History:
        """
        Trains classifiers' model on data generated by a Python generator.

        Args:
            generator: Input samples from a data generator on which to train the model.
            validation_data: Input samples from a data generator on which to evaluate the model.
            epochs: Number of epochs to train the model.
            initial_epoch: Epoch at which to start training.
            verbose: Verbosity mode.
            use_multiprocessing: Use process based threading.
            workers: Maximum number of processes.
            max_queue_size: Maximum size for the generator queue.
            callbacks: List of callbacks to apply during training.

        Returns:
            history: A `History` object.

        """
        return self.model.fit_generator(**kwargs)

    def predict_generator(self, data_generator: DataGenerator, **kwargs) -> History:
        """
        Generates predictions for the input samples from a data generator.

        Args:
            data_generator: Input samples from a data generator.
            workers: Maximum number of processes.
            use_multiprocessing: Use process based threading.
            verbose: Verbosity mode.

        Returns:
            history: A `History` object.
        """
        return self.model.predict_generator(data_generator, **kwargs)

    def summary(self):
        """ Summarizes classifiers' model."""
        self.model.summary()



from setuptools import setup, find_packages

long_description = '''
Image ATM is a one-click tool that automates the workflow of a typical
image classification pipeline in an opinionated way, this includes:

- Preprocessing and validating input images and labels
- Starting/terminating cloud instance with GPU support
- Training
- Model evaluation

Read the documentation at: https://idealo.github.io/imageatm/

Image ATM is compatible with Python 3.6 and is distributed under the Apache 2.0 license.
'''

setup(
    name='imageatm',
    version='0.1.1',
    author='Christopher Lennan, Malgorzata Adamczyk, Gunar Maiwald, Dat Tran',
    author_email='christopherlennan@gmail.com, m.adamczyk.berlin@gmail.com, gunar.maiwald@web.de, datitran@gmail.com',
    description='Image classification for everyone',
    long_description=long_description,
    license='Apache 2.0',
    install_requires=[
        'Keras==2.2.4',
        'keras-vis>=0.4.1',
        'tensorflow==1.13.1',
        'awscli',
        'Click',
        'h5py',
        'matplotlib',
        'papermill',
        'Pillow',
        'scikit-learn',
        'scipy==1.1.*',
        'tqdm',
        'yarl',
    ],
    extras_require={
        'tests': ['pytest==4.3.0', 'pytest-cov==2.6.1', 'pytest-mock', 'mock', 'mypy'],
        'docs': ['mkdocs==1.0.4', 'mkdocs-material==4.0.2'],
        'dev': ['bumpversion==0.5.3'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=('tests',)),
    entry_points={'console_scripts': ['imageatm=imageatm.client.client:cli']},
    include_package_data=True,
)



import sys
import logging
from pathlib import Path
from typing import Union


def get_logger(name: str, job_dir: Union[Path, str]) -> logging.Logger:
    if isinstance(job_dir, str):
        job_dir = Path(job_dir)

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.handlers = []

    if not logger.handlers:
        # stream handler ensures that logging events are passed to stdout
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch_formatter = logging.Formatter('%(message)s')
        ch.setFormatter(ch_formatter)
        logger.addHandler(ch)

        # file handler ensures that logging events are passed to log file
        fh = logging.FileHandler(filename=job_dir / 'logs')
        fh.setLevel(logging.DEBUG)
        fh_formatter = logging.Formatter(
            '%(asctime)s - %(module)s - %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S'
        )
        fh.setFormatter(fh_formatter)
        logger.addHandler(fh)

    return logger



from imageatm.components import DataPrep


def run_dataprep(
    image_dir: str, samples_file: str, job_dir: str, resize: bool = False, **kwargs
) -> DataPrep:
    dp = DataPrep(job_dir=job_dir, image_dir=image_dir, samples_file=samples_file, **kwargs)
    dp.run(resize=resize)

    return dp



import time
from pathlib import Path
from typing import Optional
from yarl import URL
from imageatm.utils.process import run_cmd
from imageatm.utils.logger import get_logger


class AWS:
    """Cloud provider class that allows to run training on AWS.

    Cloud instances are created and destroyed using Terraform. The instance is provisioned
    with nvidia-docker and training is run in a Docker container using the public Docker
    image [idealo/tensorflow-image-atm:1.13.1](https://hub.docker.com/r/idealo/tensorflow-image-atm/tags).

    All commands on EC2 instance will be run via SSH.

    For training the local image and job directories will be synced with S3. After training
    the trained models will be synced with S3 and the local job directory.

    Attributes:
        tf_dir: Directory with Terraform files for AWS setup.
        region: AWS region [eu-west-1, eu-central-1].
        instance_type: AWS GPU instance type [g2.\*, p2.\*, p3.\*].
        vpc_id: AWS Virtual Private Cloud ID.
        s3_bucket: AWS S3 bucket where all training files will be stored (is not created by Terraform).
        job_dir: Job directory on local system (needed for logging).
        cloud_tag: Name under which all AWS resources will be set up.
    """

    def __init__(
        self,
        tf_dir: str,
        region: str,
        instance_type: str,
        vpc_id: str,
        s3_bucket: str,
        job_dir: str,
        cloud_tag: str,
        **kwargs
    ) -> None:
        """Inits cloud component.

        Sets *remote workdir* and ensures that *s3 bucket prefix* is correct.
        """
        self.tf_dir = tf_dir
        self.region = region
        self.instance_type = instance_type
        self.vpc_id = vpc_id
        self.s3_bucket = URL(
            s3_bucket
        )  # needed for IAM setup; bucket will not be created by terraform
        self.cloud_tag = cloud_tag
        if 's3://' in str(job_dir):
            self.job_dir = URL(job_dir)
        else:
            self.job_dir = Path(job_dir).resolve()

        self.image_dir: Optional[Path] = None
        self.ssh: Optional[str] = None
        self.remote_workdir = Path('/home/ec2-user/image-atm').resolve()

        self._check_s3_prefix()

        self.logger = get_logger(__name__, Path(self.job_dir))

    def _check_s3_prefix(self):
        # ensure that s3 bucket prefix is correct
        self.s3_bucket_wo = str(self.s3_bucket).split('s3://')[-1]  # without s3:// prefix
        self.s3_bucket = URL('s3://' + self.s3_bucket_wo)

    def _set_ssh(self):
        cmd = 'cd {} && terraform output public_ip'.format(self.tf_dir)
        output = run_cmd(cmd, logger=self.logger, return_output=True)
        self.ssh = 'ssh -o StrictHostKeyChecking=no -i ~/.ssh/id_rsa ec2-user@{}'.format(output)

    def _set_remote_dirs(self):
        cmd = '{} mkdir -p {}'.format(self.ssh, self.remote_workdir)
        run_cmd(cmd, logger=self.logger)

        self.remote_image_dir = self.remote_workdir.joinpath(self.image_dir.name)
        self.remote_job_dir = self.remote_workdir.joinpath(self.job_dir.name)

    def _set_s3_dirs(self):
        if 's3://' in str(self.image_dir):
            self.s3_image_dir = self.image_dir
        else:
            self.s3_image_dir = self.s3_bucket / 'image_dirs' / self.image_dir.name

        if 's3://' in str(self.job_dir):
            self.s3_job_dir = self.job_dir
        else:
            self.s3_job_dir = self.s3_bucket / 'job_dirs' / self.job_dir.name

    def _sync_local_s3(self):
        self.logger.info('Syncing files local <> s3...')
        self._set_s3_dirs()

        # only sync if image dir is local dir
        if 's3://' not in str(self.image_dir):
            cmd = 'aws s3 sync --quiet --exclude logs {} {}'.format(
                self.image_dir, self.s3_image_dir
            )
            run_cmd(cmd, logger=self.logger)

        # only sync if job dir is local dir
        if 's3://' not in str(self.job_dir):
            cmd = 'aws s3 sync --quiet --exclude logs {} {}'.format(self.job_dir, self.s3_job_dir)
            run_cmd(cmd, logger=self.logger)

    def _sync_s3_local(self):
        self.logger.info('Syncing files s3 <> local...')
        self._set_s3_dirs()

        # only sync if job dir is local dir
        if 's3://' not in str(self.job_dir):
            cmd = 'aws s3 sync --exclude logs --quiet {} {}'.format(self.s3_job_dir, self.job_dir)
            run_cmd(cmd, logger=self.logger, level='info')

    def _sync_remote_s3(self):
        self.logger.info('Syncing files remote <> s3...')
        self._set_s3_dirs()
        self._set_remote_dirs()

        cmd = '{} aws s3 sync --exclude logs --quiet {} {}'.format(
            self.ssh, self.remote_image_dir, self.s3_image_dir
        )
        run_cmd(cmd, logger=self.logger)

        cmd = '{} aws s3 sync --exclude logs --quiet {} {}'.format(
            self.ssh, self.remote_job_dir, self.s3_job_dir
        )
        run_cmd(cmd, logger=self.logger)

    def _sync_s3_remote(self):
        self.logger.info('Syncing files s3 <> remote...')
        self._set_s3_dirs()
        self._set_remote_dirs()

        cmd = '{} aws s3 sync --exclude logs --quiet {} {}'.format(
            self.ssh, self.s3_image_dir, self.remote_image_dir
        )
        run_cmd(cmd, logger=self.logger)

        cmd = '{} aws s3 sync --exclude logs --quiet {} {}'.format(
            self.ssh, self.s3_job_dir, self.remote_job_dir
        )
        run_cmd(cmd, logger=self.logger)

    def _launch_train_container(self, **kwargs):
        self.logger.info('Launching training container...')
        # training parameters are passed to container through environment variables
        envs = ['-e {}={}'.format(key, value) for key, value in kwargs.items() if value is not None]

        cmd = (
            '{} docker run -d -v {}:$WORKDIR/image_dir -v {}:$WORKDIR/job_dir {} '
            'idealo/tensorflow-image-atm:1.13.1'
        ).format(self.ssh, self.remote_image_dir, self.remote_job_dir, ' '.join(envs))

        run_cmd(cmd, logger=self.logger)

    def _stream_docker_logs(self):
        time.sleep(5)
        cmd = '{} docker ps -l -q'.format(self.ssh)
        output = run_cmd(cmd, logger=self.logger, return_output=True)

        cmd = '{} docker logs {} --follow'.format(self.ssh, output)
        run_cmd(cmd, logger=self.logger, level='info')

    def init(self):
        """Runs Terraform initialization."""
        self.logger.info('Running terraform init...')
        cmd = 'cd {} && terraform init'.format(self.tf_dir)
        run_cmd(cmd, logger=self.logger)

    def apply(self):
        """Runs Terraform apply."""
        self.logger.info('Running terraform apply...')
        cmd = (
            'cd {} && terraform apply -auto-approve -var "region={}" -var "instance_type={}" '
            '-var "vpc_id={}" -var "s3_bucket={}" -var "name={}"'
        ).format(
            self.tf_dir,
            self.region,
            self.instance_type,
            self.vpc_id,
            self.s3_bucket_wo,
            self.cloud_tag,
        )

        run_cmd(cmd, logger=self.logger)

        self._set_ssh()

    def train(self, image_dir: str = None, job_dir: str = None, **kwargs):
        """Runs training on EC2 instance.

        The following steps will be performed in sequence:
            - syncs local image and job directory with S3
            - syncs S3 with EC2 instance
            - launches Docker training container on EC2
            - syncs EC2 with S3
            - syncs S3 with local.

        Any of the pre-trained CNNs in Keras can be used.

        Args:
            image_dir: Directory with images used for training.
            job_dir: Directory with train_samples.json, val_samples.json,
                     and class_mapping.json.
            epochs_train_dense: Number of epochs to train dense layers.
            epochs_train_all: Number of epochs to train all layers.
            learning_rate_dense: Learning rate for dense training phase.
            self.learning_rate_all: Learning rate for all training phase.
            batch_size: Number of images per batch.
            dropout_rate: Fraction set randomly.
            base_model_name: Name of pretrained CNN.
        """
        self.logger.info('Setting up remote instance...')
        if image_dir is not None:
            self.image_dir = Path(image_dir).resolve()

        if job_dir is not None:
            self.job_dir = Path(job_dir).resolve()

        self._sync_local_s3()
        self._sync_s3_remote()
        self._launch_train_container(**kwargs)
        self._stream_docker_logs()
        self._sync_remote_s3()
        self._sync_s3_local()

    def destroy(self):
        """Runs Terraform destroy."""
        self.logger.info('Running terraform destroy...')
        cmd = (
            'cd {} && terraform destroy -auto-approve -var "region={}" -var "instance_type={}" '
            '-var "vpc_id={}" -var "s3_bucket={}" -var "name={}"'
        ).format(
            self.tf_dir,
            self.region,
            self.instance_type,
            self.vpc_id,
            self.s3_bucket_wo,
            self.cloud_tag,
        )

        run_cmd(cmd, logger=self.logger)



from typing import Counter as Counter_type
from typing import Optional, List, Callable
from collections import Counter
from pathlib import Path
from imageatm.utils.images import resize_image_mp, validate_image
from imageatm.utils.io import load_json, save_json
from imageatm.utils.process import parallelise
from imageatm.utils.logger import get_logger
from sklearn.model_selection import train_test_split


MIN_CLASS_SIZE = 2
VALIDATION_SIZE = 0.15
TEST_SIZE = 0.2
PART_SIZE = 1.0


class DataPrep:
    """Prepares data for training.

    Based on a *samples file* and an *image directory* a train, validation, and test set will be created.
    Before data preparation starts samples file and image directory will be validated.

    DataPrep creates a job directory with the following set of files

    - class_mapping.json
    - train_samples.json
    - val_samples.json
    - test_samples.json

    These files are required for subsequent components (training and evaluation).


    The samples file must be in JSON format with the following keys

    [
    {
        "image_id": "image_1.jpg",
        "label": "Class 1"
    },
    {
        "image_id": "image_2.jpg",
        "label": "Class 1"
    },
    ...
    ]



    Attributes:
        image_dir: path of image directory.
        job_dir: path to job directory with samples.
        image_dir: path of image directory.
        samples_file: path to samples file.
        min_class_size: minimal number of samples per label (default 2).
        test_size: represent the proportion of the dataset to include in the test set (default 0.2).
        val_size: represent the proportion of the dataset to include in the val set (default 0.1).
        part_size: represent the proportion of the dataset to include in all sets (default 1).
    """

    def __init__(
        self,
        job_dir: str,
        image_dir: str,
        samples_file: str,
        min_class_size: int = MIN_CLASS_SIZE,
        test_size: float = TEST_SIZE,
        val_size: float = VALIDATION_SIZE,
        part_size: float = PART_SIZE,
        **kwargs
    ) -> None:
        """Inits data preparation component.

        Loads samples file. Initializes variables for further operations:
        *valid_image_ids*, *class_mapping*, *train_samples*, *val_samples*, *test_samples*.
        """
        self.job_dir = Path(job_dir).resolve()
        self.job_dir.mkdir(parents=True, exist_ok=True)

        self.image_dir = Path(image_dir)
        self.samples_file = Path(samples_file)
        self.samples_file = Path(samples_file)
        self.min_class_size = min_class_size
        self.test_size = test_size
        self.val_size = val_size
        self.part_size = part_size

        self.class_mapping: Optional[dict] = None
        self.valid_image_ids: Optional[List[str]] = None
        self.train_samples = None
        self.val_samples = None
        self.test_samples = None

        self.logger = get_logger(__name__, self.job_dir)
        self.samples = load_json(self.samples_file)

    def _get_counter(self, list_to_count: list, print_count: bool = False) -> Counter_type:
        """Retrieves samples count for each label from loaded samples and prints distribution when parameter is set."""
        counter = Counter(list_to_count)
        total = len(list_to_count)

        if print_count:
            for key, val in sorted(counter.items()):
                self.logger.info('{}: {} ({}%)'.format(key, val, round(val * 100 / total, 1)))

        return counter

    def _validate_images(self):
        """Checks if all files in *image_dir* are valid images.

        Checks if files are images with extention 'JPEG' or 'PNG' and if they are not truncated.
        self.logger.infos filenames that didn't pass the validation.

        Sets:
            self.valid_image_ids: a list of valid image.
        """
        self.logger.info('\n****** Running image validation ******\n')

        # validate images, use multiprocessing
        files = [str(i.absolute()) for i in self.image_dir.glob('*')]
        files.sort()

        results = parallelise(validate_image, files)

        valid_image_files = [j for i, j in enumerate(files) if results[i][0]]
        self.valid_image_ids = [Path(i).name for i in valid_image_files]

        # return list of invalid images to user and save them if there are more than 10
        invalid_image_files = [
            (j, str(results[i][1])) for i, j in enumerate(files) if not results[i][0]
        ]

        if invalid_image_files:
            self.logger.info('The following files are not valid image files:')
            for file_name, error_msg in invalid_image_files[:10]:
                self.logger.info('- {} ({})'.format(file_name, error_msg))
            if len(invalid_image_files) > 10:
                save_json(invalid_image_files, self.job_dir / 'invalid_image_files.json')
                self.logger.info(
                    (
                        'NOTE: More than 10 files were identified as invalid image files.\n'
                        'The full list of those files has been saved here:\n{}'.format(
                            self.job_dir / 'invalid_image_files.json'
                        )
                    )
                )

    @staticmethod
    def _validate_sample(sample: dict, valid_image_ids: list) -> bool:
        return all(
            [
                ('label' in sample),
                ('image_id' in sample),
                (sample.get('image_id') in valid_image_ids),
            ]
        )

    def _validate_samples(self):
        """Validates the samples.

        Compares samples with valid image ids. Checks if:

            - samples have 'label' and 'image_id' keys
            - there is more than 1 class
            - there are enough samples in each class

        Reassigns self.samples to valid samples.
        Sets counter of each class in samples.

        self.logger.infos:
            - samples that didn't pass the validation
            - distribution of valid samples for each label
        """
        self.logger.info('\n****** Running samples validation ******\n')

        assert self.valid_image_ids, 'Images have to be validated before samples validation.'
        valid_image_ids = set(self.valid_image_ids)  # convert to set to optimise lookup

        # exclude samples with invalid image or invalid keys
        valid_samples = []
        self.invalid_samples = []
        for sample in self.samples:
            if self._validate_sample(sample, valid_image_ids):
                valid_samples.append(sample)
            else:
                self.invalid_samples.append(sample)

        # replace self.samples with valid samples
        assert valid_samples, 'Program ended. No valid samples found.'
        self.samples = valid_samples

        # self.logger.infos invalid images to user and saves them in json if there are more than 10
        if self.invalid_samples:
            self.logger.info('The following samples were dropped:')
            for sample in self.invalid_samples[:10]:
                self.logger.info('- {}'.format(sample))

            if len(self.invalid_samples) > 10:
                self.logger.info(
                    (
                        'NOTE: {} samples were identified as invalid.\n'
                        'The full list of invalid samples will be saved in job dir.\n'.format(
                            len(self.invalid_samples)
                        )
                    )
                )

        self.logger.info('Class distribution after validation:')
        self.samples_count = self._get_counter([i['label'] for i in self.samples], print_count=True)

        # check if each class has sufficient samples
        warnings_count = 0
        for key, val in self.samples_count.items():
            if val < self.min_class_size:
                warnings_count += 1
                self.logger.info('Not enough samples for label {}'.format(key))

        assert warnings_count == 0, 'Program ended. Collect more samples.'
        assert len(self.samples_count) > 1, 'Program ended. Only one label in the dataset.'

    def _create_class_mapping(self, print_mapping: bool = True):
        """Produces a class-mapping."""
        self.class_mapping = {}
        labels = list(self.samples_count.keys())
        labels.sort()

        for i, j in enumerate(labels):
            self.class_mapping[i] = j

        if print_mapping:
            self.logger.info('Class mapping:\n{}'.format(self.class_mapping))

    def _apply_class_mapping(self):
        """Applies the class-mapping."""
        class_mapping_inv = {v: k for k, v in self.class_mapping.items()}

        samples_int = [
            {'label': class_mapping_inv[sample['label']], 'image_id': sample['image_id']}
            for sample in self.samples
        ]

        # replace self.samples with samples int
        self.samples = samples_int

    def _split_samples(self):
        """Produces stratified train, val, test sets.

        Sets:
            self.train_samples
            self.val_samples
            self.test_samples
        """
        self.logger.info('\n****** Creating train/val/test sets ******\n')

        self.train_size = 1 - (self.test_size + self.val_size)

        self.logger.info(
            'Split distribution: train: {:.2f}, val: {}, test: {:}\n'.format(
                self.train_size, self.val_size, self.test_size
            )
        )

        self.logger.info(
            'Partial split distribution: train: {:.2f}, val: {:.2f}, test: {:.2f}\n'.format(
                self.train_size * self.part_size,
                self.val_size * self.part_size,
                self.test_size * self.part_size,
            )
        )

        labels = [i['label'] for i in self.samples]  # need label list for stratification

        split_test_size = self.test_size * self.part_size
        split_train_size = (1 - self.test_size) * self.part_size

        train_samples, self.test_samples, train_labels, _ = train_test_split(
            self.samples,
            labels,
            test_size=split_test_size,
            train_size=split_train_size,
            shuffle=True,
            random_state=10207,
            stratify=labels,
        )

        split_test_size = self.val_size / (1 - self.test_size)
        self.train_samples, self.val_samples, _, _ = train_test_split(
            train_samples,
            train_labels,
            test_size=split_test_size,
            shuffle=True,
            random_state=10207,
            stratify=train_labels,
        )

        self.logger.info('Train set:')
        self._get_counter([i['label'] for i in self.train_samples], print_count=True)

        self.logger.info('Val set:')
        self._get_counter([i['label'] for i in self.val_samples], print_count=True)

        self.logger.info('Test set:')
        self._get_counter([i['label'] for i in self.test_samples], print_count=True)

    def _resize_images(self, resize_image_mp: Callable = resize_image_mp):
        self.logger.info('\n****** Resizing images ******\n')
        new_image_dir = '_'.join([str(self.image_dir), 'resized'])
        Path(new_image_dir).resolve().mkdir(parents=True, exist_ok=True)

        args = [(self.image_dir, new_image_dir, i['image_id']) for i in self.samples]
        parallelise(resize_image_mp, args)
        self.logger.info(
            'Stored {} resized images under {}'.format(len(self.samples), new_image_dir)
        )

        self.image_dir = Path(new_image_dir)
        self.logger.info('Changed image directory to {}'.format(self.image_dir))

    def _save_files(self):
        save_json(self.train_samples, self.job_dir / 'train_samples.json')
        save_json(self.val_samples, self.job_dir / 'val_samples.json')
        save_json(self.test_samples, self.job_dir / 'test_samples.json')
        save_json(self.class_mapping, self.job_dir / 'class_mapping.json')

        if len(self.invalid_samples) > 10:
            save_json(self.invalid_samples, self.job_dir / 'invalid_samples.json')
            self.logger.info(
                (
                    'NOTE: More than 10 samples were identified as invalid.\n'
                    'The full list of invalid samples has been saved here:\n{}'.format(
                        self.job_dir / 'invalid_samples.json'
                    )
                )
            )

    def run(self, resize: bool = False):
        """Executes all steps of data preparation.

            - Validates samples and images
            - Creates class-mapping (string to integer)
            - Applies class-mapping on samples
            - Splits sample into train, validation and test sets
            - Resizes images
            - Saves files (class-mapping, train-, validation- and test-set)

        Args:
            resize: boolean (creates a subfolder of resized images, default False).
        """
        self._validate_images()
        self._validate_samples()

        self._create_class_mapping()
        self._apply_class_mapping()
        self._split_samples()

        if resize:
            self._resize_images()

        self._save_files()
